#!/usr/bin/env python3
"""Example: Create new image from container."""

import sys

import podman


def print_history(details):
    """Format history data from an image, in a table."""
    for i, r in enumerate(details):
        print(
            '{}: {} {} {}'.format(i, r.id[:12],
                                  podman.datetime_format(r.created), r.tags),
            sep='\n')
    print("-" * 25)


print('{}\n'.format(__doc__))

with podman.Client() as client:
    ctnr = next(
        (c for c in client.containers.list() if 'alpine' in c['image']), None)

    if ctnr:
        print_history(client.images.get(ctnr.imageid).history())

        # Make changes as we save the container to a new image
        id = ctnr.commit('alpine-ash', changes=['CMD=/bin/ash'])
        print_history(client.images.get(id).history())
    else:
        print('Unable to find "alpine" container.', file=sys.stderr)



"""Remote client command for attaching to a container."""
import sys

import podman
from pypodman.lib import AbstractActionBase


class Attach(AbstractActionBase):
    """Class for attaching to a running container."""

    @classmethod
    def subparser(cls, parent):
        """Add Attach command to parent parser."""
        parser = parent.add_parser('attach', help='attach to container')
        parser.add_argument(
            '--image',
            help='image to instantiate and attach to',
        )
        parser.add_argument(
            'command',
            nargs='*',
            help='image to instantiate and attach to',
        )
        parser.set_defaults(class_=cls, method='attach')

    def __init__(self, args):
        """Construct Attach class."""
        super().__init__(args)
        if not args.image:
            raise ValueError('You must supply one image id'
                             ' or name to be attached.')

    def attach(self):
        """Attach to instantiated image."""
        args = {
            'detach': True,
            'tty': True,
        }
        if self._args.command:
            args['command'] = self._args.command

        try:
            try:
                ident = self.client.images.pull(self._args.image)
                img = self.client.images.get(ident)
            except podman.ImageNotFound as e:
                sys.stdout.flush()
                print(
                    'Image {} not found.'.format(e.name),
                    file=sys.stderr,
                    flush=True)
                return 1

            ctnr = img.create(**args)
            ctnr.attach(eot=4)

            try:
                ctnr.start()
                print()
            except (BrokenPipeError, KeyboardInterrupt):
                print('\nContainer disconnected.')
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1



"""Parse configuration while building subcommands."""
import argparse
import getpass
import inspect
import logging
import os
import shutil
import sys
from contextlib import suppress
from pathlib import Path

import pkg_resources
import pytoml

from .parser_actions import PathAction, PositiveIntAction

# TODO: setup.py and obtain __version__ from rpm.spec
try:
    __version__ = pkg_resources.get_distribution('pypodman').version
except Exception:  # pylint: disable=broad-except
    __version__ = '0.0.0'


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Set help width to screen size."""

    def __init__(self, *args, **kwargs):
        """Construct HelpFormatter using screen width."""
        if 'width' not in kwargs:
            try:
                size = shutil.get_terminal_size()
                kwargs['width'] = size.columns
            except Exception:  # pylint: disable=broad-except
                kwargs['width'] = 80

        super().__init__(*args, **kwargs)


class PodmanArgumentParser(argparse.ArgumentParser):
    """Default remote podman configuration."""

    def __init__(self, **kwargs):
        """Construct the parser."""
        kwargs['add_help'] = True
        kwargs['description'] = ('Portable and simple management'
                                 ' tool for containers and images')
        kwargs['formatter_class'] = HelpFormatter

        super().__init__(**kwargs)

    def initialize_parser(self):
        """Initialize parser without causing recursion meltdown."""
        self.add_argument(
            '--version',
            action='version',
            version='%(prog)s v. ' + __version__)
        self.add_argument(
            '--log-level',
            choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
            default='WARNING',
            type=str.upper,
            help='set logging level for events. (default: %(default)s)',
        )
        self.add_argument(
            '--run-dir',
            metavar='DIRECTORY',
            help=('directory to place local socket bindings.'
                  ' (default: XDG_RUNTIME_DIR/pypodman)'))
        self.add_argument(
            '--username',
            '-l',
            help='Authenicating user on remote host. (default: {})'.format(
                getpass.getuser()))
        self.add_argument(
            '--host', help='name of remote host. (default: None)')
        self.add_argument(
            '--port',
            '-p',
            action=PositiveIntAction,
            help='port for ssh tunnel to remote host. (default: 22)')
        self.add_argument(
            '--remote-socket-path',
            metavar='PATH',
            help=('path of podman socket on remote host'
                  ' (default: /run/podman/io.podman)'))
        self.add_argument(
            '--identity-file',
            '-i',
            action=PathAction,
            help='path to ssh identity file. (default: ~user/.ssh/id_dsa)')
        self.add_argument(
            '--config-home',
            metavar='DIRECTORY',
            action=PathAction,
            help=('home of configuration "pypodman.conf".'
                  ' (default: XDG_CONFIG_HOME/pypodman)'))

        actions_parser = self.add_subparsers(
            dest='subparser_name', help='commands')

        # import buried here to prevent import loops
        import pypodman.lib.actions  # pylint: disable=cyclic-import
        assert pypodman.lib.actions

        # pull in plugin(s) code for each subcommand
        for name, obj in inspect.getmembers(
                sys.modules['pypodman.lib.actions'],
                predicate=inspect.isclass):
            if hasattr(obj, 'subparser'):
                try:
                    obj.subparser(actions_parser)
                except NameError as e:
                    logging.critical(e)
                    logging.warning(
                        'See subparser configuration for Class "%s"', name)
                    sys.exit(3)

    def parse_args(self, args=None, namespace=None):
        """Parse command line arguments, backed by env var and config_file."""
        self.initialize_parser()
        cooked = super().parse_args(args, namespace)
        return self.resolve_configuration(cooked)

    def resolve_configuration(self, args):
        """Find and fill in any arguments not passed on command line."""
        args.xdg_runtime_dir = os.environ.get('XDG_RUNTIME_DIR', '/tmp')
        args.xdg_config_home = os.environ.get('XDG_CONFIG_HOME',
                                              os.path.expanduser('~/.config'))
        args.xdg_config_dirs = os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg')

        # Configuration file(s) are optional,
        #   required arguments may be provided elsewhere
        config = {'default': {}}
        dirs = args.xdg_config_dirs.split(':')
        dirs.extend((args.xdg_config_home, args.config_home))
        for dir_ in dirs:
            if dir_ is None:
                continue
            with suppress(OSError):
                cnf = Path(dir_, 'pypodman', 'pypodman.conf')
                with cnf.open() as stream:
                    config.update(pytoml.load(stream))

        def reqattr(name, value):
            """Raise an error if value is unset."""
            if value:
                setattr(args, name, value)
                return value
            return self.error(
                'Required argument "{}" is not configured.'.format(name))

        reqattr(
            'run_dir',
            getattr(args, 'run_dir')
            or os.environ.get('RUN_DIR')
            or config['default'].get('run_dir')
            or Path(args.xdg_runtime_dir, 'pypodman')
        )   # yapf: disable

        setattr(
            args,
            'host',
            getattr(args, 'host')
            or os.environ.get('HOST')
            or config['default'].get('host')
        )   # yapf:disable

        reqattr(
            'username',
            getattr(args, 'username')
            or os.environ.get('USER')
            or os.environ.get('LOGNAME')
            or config['default'].get('username')
            or getpass.getuser()
        )   # yapf:disable

        reqattr(
            'port',
            getattr(args, 'port')
            or os.environ.get('PORT')
            or config['default'].get('port', None)
            or 22
        )   # yapf:disable

        reqattr(
            'remote_socket_path',
            getattr(args, 'remote_socket_path')
            or os.environ.get('REMOTE_SOCKET_PATH')
            or config['default'].get('remote_socket_path')
            or '/run/podman/io.podman'
        )   # yapf:disable

        reqattr(
            'log_level',
            getattr(args, 'log_level')
            or os.environ.get('LOG_LEVEL')
            or config['default'].get('log_level')
            or logging.WARNING
        )  # yapf:disable

        setattr(
            args,
            'identity_file',
            getattr(args, 'identity_file')
            or os.environ.get('IDENTITY_FILE')
            or config['default'].get('identity_file')
            or os.path.expanduser('~{}/.ssh/id_dsa'.format(args.username))
        )   # yapf:disable

        if not os.path.isfile(args.identity_file):
            args.identity_file = None

        if args.host:
            args.local_socket_path = Path(args.run_dir, 'podman.socket')
        else:
            args.local_socket_path = args.remote_socket_path

        args.local_uri = 'unix:{}'.format(args.local_socket_path)

        if args.host:
            components = ['ssh://', args.username, '@', args.host]
            if args.port:
                components.extend((':', str(args.port)))
            components.append(args.remote_socket_path)

            args.remote_uri = ''.join(components)
        return args

    def exit(self, status=0, message=None):
        """Capture message and route to logger."""
        if message:
            log = logging.info if status == 0 else logging.error
            log(message)
        super().exit(status)

    def error(self, message):
        """Capture message and route to logger."""
        logging.error('%s: %s', self.prog, message)
        logging.error("Try '%s --help' for more information.", self.prog)
        super().exit(2)



"""Report Manager."""
import sys
from collections import namedtuple


class ReportColumn(namedtuple('ReportColumn', 'key display width default')):
    """Hold attributes of output column."""

    __slots__ = ()

    def __new__(cls, key, display, width, default=None):
        """Add defaults for attributes."""
        return super(ReportColumn, cls).__new__(cls, key, display, width,
                                                default)


class Report():
    """Report Manager."""

    def __init__(self, columns, heading=True, epilog=None, file=sys.stdout):
        """Construct Report.

        columns is a mapping for named fields to column headings.
        headers True prints headers on table.
        epilog will be printed when the report context is closed.
        """
        self._columns = columns
        self._file = file
        self._heading = heading
        self.epilog = epilog
        self._format = None

    def row(self, **fields):
        """Print row for report."""
        if self._heading:
            hdrs = {k: v.display for (k, v) in self._columns.items()}
            print(self._format.format(**hdrs), flush=True, file=self._file)
            self._heading = False
        fields = {k: str(v) for k, v in fields.items()}
        print(self._format.format(**fields))

    def __enter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Leave Report context and print epilog if provided."""
        if self.epilog:
            print(self.epilog, flush=True, file=self._file)

    def layout(self, iterable, keys, truncate=True):
        """Use data and headings build format for table to fit."""
        fmt = []

        for key in keys:
            slice_ = [str(i.get(key, '')) for i in iterable]
            data_len = len(max(slice_, key=len))

            info = self._columns.get(key,
                                     ReportColumn(key, key.upper(), data_len))
            display_len = max(data_len, len(info.display))
            if truncate and info.width != 0:
                display_len = info.width

            fmt.append('{{{0}:{1}.{1}}}'.format(key, display_len))
        self._format = ' '.join(fmt)



#!/usr/bin/env python3
"""Example: Pull Fedora and inspect image and container."""

import podman

print('{}\n'.format(__doc__))

with podman.Client() as client:
    id = client.images.pull('registry.fedoraproject.org/fedora:28')
    img = client.images.get(id)
    print(img.inspect())

    cntr = img.create()
    print(cntr.inspect())

    cntr.remove()



from __future__ import absolute_import

import unittest

from report import Report, ReportColumn


class TestReport(unittest.TestCase):
    def setUp(self):
        pass

    def test_report_column(self):
        rc = ReportColumn('k', 'v', 3)
        self.assertEqual(rc.key, 'k')
        self.assertEqual(rc.display, 'v')
        self.assertEqual(rc.width, 3)
        self.assertIsNone(rc.default)

        rc = ReportColumn('k', 'v', 3, 'd')
        self.assertEqual(rc.key, 'k')
        self.assertEqual(rc.display, 'v')
        self.assertEqual(rc.width, 3)
        self.assertEqual(rc.default, 'd')



"""
Supplimental argparse.Action converters and validaters.

The constructors are very verbose but remain for IDE support.
"""
import argparse
import os

# API defined by argparse.Action shut up pylint
# pragma pylint: disable=redefined-builtin
# pragma pylint: disable=too-few-public-methods
# pragma pylint: disable=too-many-arguments


class BooleanValidate():
    """Validate value is boolean string."""

    def __call__(self, value):
        """Return True, False or raise ValueError."""
        val = value.capitalize()
        if val == 'False':
            return False
        elif val == 'True':
            return True
        else:
            raise ValueError('"{}" is not True or False'.format(value))


class BooleanAction(argparse.Action):
    """Convert and validate bool argument."""

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=None,
                 choices=('True', 'False'),
                 required=False,
                 help=None,
                 metavar='{True,False}'):
        """Create BooleanAction object."""
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        """Convert and Validate input."""
        try:
            val = BooleanValidate()(values)
        except ValueError:
            parser.error('{} must be True or False.'.format(self.dest))
        else:
            setattr(namespace, self.dest, val)


class UnitAction(argparse.Action):
    """Validate number given is positive integer, with optional suffix."""

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=None,
                 choices=None,
                 required=False,
                 help=None,
                 metavar='UNIT'):
        """Create UnitAction object."""
        help = (help or metavar or dest
                ) + ' (format: <number>[<unit>], where unit = b, k, m or g)'
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        """Validate input as a UNIT."""
        try:
            val = int(values)
        except ValueError:
            if not values[:-1].isdigit():
                msg = ('{} must be a positive integer,'
                       ' with optional suffix').format(self.dest)
                parser.error(msg)
            if not values[-1] in ('b', 'k', 'm', 'g'):
                msg = '{} only supports suffices of: b, k, m, g'.format(
                    self.dest)
                parser.error(msg)
        else:
            if val <= 0:
                msg = '{} must be a positive integer'.format(self.dest)
                parser.error(msg)

        setattr(namespace, self.dest, values)


class PositiveIntAction(argparse.Action):
    """Validate number given is positive integer."""

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=int,
                 choices=None,
                 required=False,
                 help=None,
                 metavar=None):
        """Create PositiveIntAction object."""
        self.message = '{} must be a positive integer'.format(dest)
        help = help or self.message

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=int,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        """Validate input."""
        if values > 0:
            setattr(namespace, self.dest, values)
            return

        parser.error(self.message)


class PathAction(argparse.Action):
    """Expand user- and relative-paths."""

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=None,
                 choices=None,
                 required=False,
                 help=None,
                 metavar='PATH'):
        """Create PathAction object."""
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        """Resolve full path value on local filesystem."""
        setattr(namespace, self.dest,
                os.path.abspath(os.path.expanduser(values)))



"""Remote client command for pushing image elsewhere."""
import sys

import podman
from pypodman.lib import AbstractActionBase


class Push(AbstractActionBase):
    """Class for pushing images to repository."""

    @classmethod
    def subparser(cls, parent):
        """Add Push command to parent parser."""
        parser = parent.add_parser(
            'push',
            help='push image elsewhere',
        )
        parser.add_argument(
            '--tlsverify',
            action='store_true',
            default=True,
            help='Require HTTPS and verify certificates when'
            ' contacting registries (default: %(default)s)')
        parser.add_argument(
            'image', nargs=1, help='name or id of image to push')
        parser.add_argument(
            'tag',
            nargs=1,
            help='destination image id',
        )
        parser.set_defaults(class_=cls, method='push')

    def __init__(self, args):
        """Construct Push class."""
        super().__init__(args)

    def pull(self):
        """Store image elsewhere."""
        try:
            try:
                img = self.client.images.get(self._args.image[0])
            except podman.ImageNotFound as e:
                sys.stdout.flush()
                print(
                    'Image {} not found.'.format(e.name),
                    file=sys.stderr,
                    flush=True)
            else:
                img.push(self._args.tag[0], tlsverify=self._args.tlsverify)
                print(self._args.image[0])
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)



"""Base for podman tests."""
import contextlib
import functools
import itertools
import os
import subprocess
import time
import unittest

from varlink import VarlinkError

MethodNotImplemented = 'org.varlink.service.MethodNotImplemented'


class PodmanTestCase(unittest.TestCase):
    """Hide the sausage making of initializing storage."""

    @classmethod
    def setUpClass(cls):
        """Fixture to setup podman test case."""
        if hasattr(PodmanTestCase, 'alpine_process'):
            PodmanTestCase.tearDownClass()

        def run_cmd(*args):
            cmd = list(itertools.chain(*args))
            try:
                pid = subprocess.Popen(
                    cmd,
                    close_fds=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
                out, err = pid.communicate()
            except OSError as e:
                print('{}: {}({})'.format(cmd, e.strerror, e.returncode))
            except ValueError as e:
                print('{}: {}'.format(cmd, e.message))
                raise
            else:
                return out.strip()

        tmpdir = os.environ.get('TMPDIR', '/tmp')
        podman_args = [
            '--storage-driver=vfs',
            '--cgroup-manager=cgroupfs',
            '--root={}/crio'.format(tmpdir),
            '--runroot={}/crio-run'.format(tmpdir),
            '--cni-config-dir={}/cni/net.d'.format(tmpdir),
        ]

        run_podman = functools.partial(run_cmd, ['podman'], podman_args)

        id = run_podman(['pull', 'alpine'])
        setattr(PodmanTestCase, 'alpine_id', id)

        run_podman(['pull', 'busybox'])
        run_podman(['images'])

        run_cmd(['rm', '-f', '{}/alpine_gold.tar'.format(tmpdir)])
        run_podman([
            'save', '--output', '{}/alpine_gold.tar'.format(tmpdir), 'alpine'
        ])

        PodmanTestCase.alpine_log = open(
            os.path.join('/tmp/', 'alpine.log'), 'w')

        cmd = ['podman']
        cmd.extend(podman_args)
        # cmd.extend(['run', '-d', 'alpine', 'sleep', '500'])
        cmd.extend(['run', '-dt', 'alpine', '/bin/sh'])
        PodmanTestCase.alpine_process = subprocess.Popen(
            cmd,
            stdout=PodmanTestCase.alpine_log,
            stderr=subprocess.STDOUT,
        )

        PodmanTestCase.busybox_log = open(
            os.path.join('/tmp/', 'busybox.log'), 'w')

        cmd = ['podman']
        cmd.extend(podman_args)
        cmd.extend(['create', 'busybox'])
        PodmanTestCase.busybox_process = subprocess.Popen(
            cmd,
            stdout=PodmanTestCase.busybox_log,
            stderr=subprocess.STDOUT,
        )
        # give podman time to start ctnr
        time.sleep(2)

        # Close our handle of file
        PodmanTestCase.alpine_log.close()
        PodmanTestCase.busybox_log.close()

    @classmethod
    def tearDownClass(cls):
        """Fixture to clean up after podman unittest."""
        try:
            PodmanTestCase.alpine_process.kill()
            assert 0 == PodmanTestCase.alpine_process.wait(500)
            delattr(PodmanTestCase, 'alpine_process')

            PodmanTestCase.busybox_process.kill()
            assert 0 == PodmanTestCase.busybox_process.wait(500)
        except Exception as e:
            print('Exception: {}'.format(e))
            raise

    @contextlib.contextmanager
    def assertRaisesNotImplemented(self):
        """Sugar for unimplemented varlink methods."""
        with self.assertRaisesRegex(VarlinkError, MethodNotImplemented):
            yield



"""Remote client command for retrieving container logs."""
import argparse
import logging
import sys
from collections import deque

import podman
from pypodman.lib import AbstractActionBase, PositiveIntAction


class Logs(AbstractActionBase):
    """Class for retrieving logs from container."""

    @classmethod
    def subparser(cls, parent):
        """Add Logs command to parent parser."""
        parser = parent.add_parser('logs', help='retrieve logs from container')
        parser.add_argument(
            '--tail',
            metavar='LINES',
            action=PositiveIntAction,
            help='Output the specified number of LINES at the end of the logs')
        parser.add_argument(
            'container',
            nargs=1,
            help='retrieve container logs',
        )
        parser.set_defaults(class_=cls, method='logs')

    def __init__(self, args):
        """Construct Logs class."""
        super().__init__(args)

    def logs(self):
        """Retrieve logs from containers."""
        try:
            ident = self._args.container[0]
            try:
                logging.debug('Get container "%s" logs', ident)
                ctnr = self.client.containers.get(ident)
            except podman.ContainerNotFound as e:
                sys.stdout.flush()
                print(
                    'Container "{}" not found'.format(e.name),
                    file=sys.stderr,
                    flush=True)
            else:
                if self._args.tail:
                    logs = iter(deque(ctnr.logs(), maxlen=self._args.tail))
                else:
                    logs = ctnr.logs()

                for line in logs:
                    sys.stdout.write(line)
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1



"""Remote client command for creating container from image."""
import sys

import podman
from pypodman.lib import AbstractActionBase

from ._create_args import CreateArguments


class Create(AbstractActionBase):
    """Class for creating container from image."""

    @classmethod
    def subparser(cls, parent):
        """Add Create command to parent parser."""
        parser = parent.add_parser(
            'create', help='create container from image')

        CreateArguments.add_arguments(parser)

        parser.add_argument('image', nargs=1, help='source image id')
        parser.add_argument(
            'command',
            nargs='*',
            help='command and args to run.',
        )
        parser.set_defaults(class_=cls, method='create')

    def __init__(self, args):
        """Construct Create class."""
        super().__init__(args)

        # image id used only on client
        del self.opts['image']

    def create(self):
        """Create container."""
        try:
            for ident in self._args.image:
                try:
                    img = self.client.images.get(ident)
                    img.container(**self.opts)
                    print(ident)
                except podman.ImageNotFound as e:
                    sys.stdout.flush()
                    print(
                        'Image {} not found.'.format(e.name),
                        file=sys.stderr,
                        flush=True)
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)



import os
import unittest

import podman
import varlink

ident = None
pod = None


class TestPodsNoCtnrs(unittest.TestCase):
    def setUp(self):
        self.tmpdir = os.environ['TMPDIR']
        self.host = os.environ['PODMAN_HOST']

        self.pclient = podman.Client(self.host)

    def test_010_create(self):
        global ident

        actual = self.pclient.pods.create('pod0')
        self.assertIsNotNone(actual)
        ident = actual.id

    def test_015_list(self):
        global ident, pod

        actual = next(self.pclient.pods.list())
        self.assertEqual('pod0', actual.name)
        self.assertEqual(ident, actual.id)
        self.assertEqual('Created', actual.status)
        self.assertEqual('0', actual.numberofcontainers)
        self.assertFalse(actual.containersinfo)
        pod = actual

    def test_020_get(self):
        global ident, pod

        actual = self.pclient.pods.get(pod.id)
        self.assertEqual('pod0', actual.name)
        self.assertEqual(ident, actual.id)
        self.assertEqual('Created', actual.status)
        self.assertEqual('0', actual.numberofcontainers)
        self.assertFalse(actual.containersinfo)

    def test_025_inspect(self):
        global ident, pod

        details = pod.inspect()
        self.assertEqual(ident, details.id)
        self.assertEqual('pod0', details.config['name'])
        self.assertIsNone(details.containers)

    def test_030_ident_no_ctnrs(self):
        global ident, pod

        actual = pod.kill()
        self.assertEqual(pod, actual)

        actual = pod.pause()
        self.assertEqual(pod, actual)

        actual = pod.unpause()
        self.assertEqual(pod, actual)

        actual = pod.stop()
        self.assertEqual(pod, actual)

    def test_045_raises_no_ctnrs(self):
        global ident, pod

        with self.assertRaises(podman.NoContainersInPod):
            pod.start()

        with self.assertRaises(podman.NoContainersInPod):
            pod.restart()

        with self.assertRaises(podman.NoContainerRunning):
            next(pod.stats())

        with self.assertRaises(varlink.error.MethodNotImplemented):
            pod.top()

        with self.assertRaises(varlink.error.MethodNotImplemented):
            pod.wait()

    def test_999_remove(self):
        global ident, pod

        actual = pod.remove()
        self.assertEqual(ident, actual)

        with self.assertRaises(StopIteration):
            next(self.pclient.pods.list())



"""Remote client command for inspecting podman objects."""
import json
import logging
import sys

import podman
from pypodman.lib import AbstractActionBase


class Inspect(AbstractActionBase):
    """Class for inspecting podman objects."""

    @classmethod
    def subparser(cls, parent):
        """Add Inspect command to parent parser."""
        parser = parent.add_parser('inspect', help='inspect objects')
        parser.add_argument(
            '--type',
            '-t',
            choices=('all', 'container', 'image'),
            default='all',
            type=str.lower,
            help='Type of object to inspect',
        )
        parser.add_argument(
            'size',
            action='store_true',
            default=True,
            help='Display the total file size if the type is a container.'
            ' Always True.')
        parser.add_argument(
            'objects',
            nargs='+',
            help='objects to inspect',
        )
        parser.set_defaults(class_=cls, method='inspect')

    def __init__(self, args):
        """Construct Inspect class."""
        super().__init__(args)

    def _get_container(self, ident):
        try:
            logging.debug("Getting container %s", ident)
            ctnr = self.client.containers.get(ident)
        except podman.ContainerNotFound:
            pass
        else:
            return ctnr.inspect()

    def _get_image(self, ident):
        try:
            logging.debug("Getting image %s", ident)
            img = self.client.images.get(ident)
        except podman.ImageNotFound:
            pass
        else:
            return img.inspect()

    def inspect(self):
        """Inspect provided podman objects."""
        output = {}
        try:
            for ident in self._args.objects:
                obj = None

                if self._args.type in ('all', 'container'):
                    obj = self._get_container(ident)
                if obj is None and self._args.type in ('all', 'image'):
                    obj = self._get_image(ident)

                if obj is None:
                    if self._args.type == 'container':
                        msg = 'Container "{}" not found'.format(ident)
                    elif self._args.type == 'image':
                        msg = 'Image "{}" not found'.format(ident)
                    else:
                        msg = 'Object "{}" not found'.format(ident)
                    print(msg, file=sys.stderr, flush=True)
                else:
                    output.update(obj._asdict())
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1
        else:
            print(json.dumps(output, indent=2))



#!/usr/bin/env python3
"""Example: Run top on Alpine container."""

import podman

print('{}\n'.format(__doc__))

with podman.Client() as client:
    id = client.images.pull('alpine:latest')
    img = client.images.get(id)
    cntr = img.create(detach=True, tty=True, command=['/usr/bin/top'])
    cntr.attach(eot=4)

    try:
        cntr.start()
        print()
    except (BrokenPipeError, KeyboardInterrupt):
        print('\nContainer disconnected.')



"""Remote client command for retrieving mounts from containers."""
import sys
from collections import OrderedDict

import podman
from pypodman.lib import AbstractActionBase, Report, ReportColumn


class Mount(AbstractActionBase):
    """Class for retrieving mounts from container."""

    @classmethod
    def subparser(cls, parent):
        """Add mount command to parent parser."""
        parser = parent.add_parser(
            'mount', help='retrieve mounts from containers.')
        super().subparser(parser)
        parser.add_argument(
            'containers',
            nargs='*',
            help='containers to list ports',
        )
        parser.set_defaults(class_=cls, method='mount')

    def __init__(self, args):
        """Construct Mount class."""
        super().__init__(args)

        self.columns = OrderedDict({
            'id':
            ReportColumn('id', 'CONTAINER ID', 14),
            'destination':
            ReportColumn('destination', 'DESTINATION', 0)
        })

    def mount(self):
        """Retrieve mounts from containers."""
        try:
            ctnrs = []
            if not self._args.containers:
                ctnrs = self.client.containers.list()
            else:
                for ident in self._args.containers:
                    try:
                        ctnrs.append(self.client.containers.get(ident))
                    except podman.ContainerNotFound as e:
                        sys.stdout.flush()
                        print(
                            'Container "{}" not found'.format(e.name),
                            file=sys.stderr,
                            flush=True)

        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1

        if not ctnrs:
            print(
                'Unable to find any containers.', file=sys.stderr, flush=True)
            return 1

        rows = list()
        for ctnr in ctnrs:
            details = ctnr.inspect()
            rows.append({
                'id': ctnr.id,
                'destination': details.graphdriver['data']['mergeddir']
            })

        with Report(self.columns, heading=self._args.heading) as report:
            report.layout(
                rows, self.columns.keys(), truncate=self._args.truncate)
            for row in rows:
                report.row(**row)



"""Remote client commands dealing with containers."""
import operator
from collections import OrderedDict

import humanize

import podman
from pypodman.lib import AbstractActionBase, Report, ReportColumn


class Ps(AbstractActionBase):
    """Class for Container manipulation."""

    @classmethod
    def subparser(cls, parent):
        """Add Images command to parent parser."""
        parser = parent.add_parser('ps', help='list containers')
        super().subparser(parser)
        parser.add_argument(
            '--sort',
            choices=('createdat', 'id', 'image', 'names', 'runningfor', 'size',
                     'status'),
            default='createdat',
            type=str.lower,
            help=('Change sort ordered of displayed containers.'
                  ' (default: %(default)s)'))
        parser.set_defaults(class_=cls, method='list')

    def __init__(self, args):
        """Construct Ps class."""
        super().__init__(args)

        self.columns = OrderedDict({
            'id':
            ReportColumn('id', 'CONTAINER ID', 14),
            'image':
            ReportColumn('image', 'IMAGE', 30),
            'command':
            ReportColumn('column', 'COMMAND', 20),
            'createdat':
            ReportColumn('createdat', 'CREATED', 12),
            'status':
            ReportColumn('status', 'STATUS', 10),
            'ports':
            ReportColumn('ports', 'PORTS', 0),
            'names':
            ReportColumn('names', 'NAMES', 18)
        })

    def list(self):
        """List containers."""
        # TODO: Verify sorting on dates and size
        ctnrs = sorted(
            self.client.containers.list(),
            key=operator.attrgetter(self._args.sort))
        if not ctnrs:
            return

        rows = list()
        for ctnr in ctnrs:
            fields = dict(ctnr)
            fields.update({
                'command':
                ' '.join(ctnr.command),
                'createdat':
                humanize.naturaldate(podman.datetime_parse(ctnr.createdat)),
            })

            if self._args.truncate:
                fields.update({'image': ctnr.image[-30:]})
            rows.append(fields)

        with Report(self.columns, heading=self._args.heading) as report:
            report.layout(
                rows, self.columns.keys(), truncate=self._args.truncate)
            for row in rows:
                report.row(**row)



import os
import signal
import unittest
from pathlib import Path
from test.podman_testcase import PodmanTestCase
from test.retry_decorator import retry

import podman


class TestContainers(PodmanTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        self.tmpdir = os.environ['TMPDIR']
        self.host = os.environ['PODMAN_HOST']

        self.pclient = podman.Client(self.host)
        self.loadCache()

    def tearDown(self):
        pass

    def loadCache(self):
        self.containers = list(self.pclient.containers.list())

        self.alpine_ctnr = next(
            iter([c for c in self.containers if 'alpine' in c['image']] or []),
            None)

        if self.alpine_ctnr and self.alpine_ctnr.status != 'running':
            self.alpine_ctnr.start()

    def test_list(self):
        self.assertGreaterEqual(len(self.containers), 2)
        self.assertIsNotNone(self.alpine_ctnr)
        self.assertIn('alpine', self.alpine_ctnr.image)

    def test_delete_stopped(self):
        before = len(self.containers)

        self.alpine_ctnr.stop()
        target = self.alpine_ctnr.id
        actual = self.pclient.containers.delete_stopped()
        self.assertIn(target, actual)

        self.loadCache()
        after = len(self.containers)

        self.assertLess(after, before)
        TestContainers.setUpClass()

    def test_get(self):
        actual = self.pclient.containers.get(self.alpine_ctnr.id)
        for k in ['id', 'status', 'ports']:
            self.assertEqual(actual[k], self.alpine_ctnr[k])

        with self.assertRaises(podman.ContainerNotFound):
            self.pclient.containers.get("bozo")

    def test_attach(self):
        # StringIO does not support fileno() so we had to go old school
        input = os.path.join(self.tmpdir, 'test_attach.stdin')
        output = os.path.join(self.tmpdir, 'test_attach.stdout')

        with open(input, 'w+') as mock_in, open(output, 'w+') as mock_out:
            # double quote is indeed in the expected place
            mock_in.write('echo H"ello, World"; exit\n')
            mock_in.seek(0, 0)

            ctnr = self.pclient.images.get(self.alpine_ctnr.image).container(
                detach=True, tty=True)
            ctnr.attach(stdin=mock_in.fileno(), stdout=mock_out.fileno())
            ctnr.start()

            mock_out.flush()
            mock_out.seek(0, 0)
            output = mock_out.read()
            self.assertIn('Hello', output)

            ctnr.remove(force=True)

    def test_processes(self):
        actual = list(self.alpine_ctnr.processes())
        self.assertGreaterEqual(len(actual), 2)

    def test_start_stop_wait(self):
        ctnr = self.alpine_ctnr.stop()
        self.assertFalse(ctnr['running'])

        ctnr.start()
        self.assertTrue(ctnr.running)

        ctnr.stop()
        self.assertFalse(ctnr['containerrunning'])

        actual = ctnr.wait()
        self.assertGreaterEqual(actual, 0)

    def test_changes(self):
        actual = self.alpine_ctnr.changes()

        self.assertListEqual(
            sorted(['changed', 'added', 'deleted']), sorted(
                list(actual.keys())))

        # TODO: brittle, depends on knowing history of ctnr
        self.assertGreaterEqual(len(actual['changed']), 0)
        self.assertGreaterEqual(len(actual['added']), 0)
        self.assertEqual(len(actual['deleted']), 0)

    def test_kill(self):
        self.assertTrue(self.alpine_ctnr.running)
        ctnr = self.alpine_ctnr.kill(signal.SIGKILL)
        self.assertFalse(ctnr.running)

    def test_inspect(self):
        actual = self.alpine_ctnr.inspect()
        self.assertEqual(actual.id, self.alpine_ctnr.id)
        # TODO: Datetime values from inspect missing offset in CI instance
        # self.assertEqual(
        # datetime_parse(actual.created),
        # datetime_parse(self.alpine_ctnr.createdat))

    def test_export(self):
        target = os.path.join(self.tmpdir, 'alpine_export_ctnr.tar')

        actual = self.alpine_ctnr.export(target)
        self.assertEqual(actual, target)
        self.assertTrue(os.path.isfile(target))
        self.assertGreater(os.path.getsize(target), 0)

    def test_commit(self):
        # TODO: Test for STOPSIGNAL when supported by OCI
        # TODO: Test for message when supported by OCI
        details = self.pclient.images.get(self.alpine_ctnr.image).inspect()
        changes = ['ENV=' + i for i in details.containerconfig['env']]
        changes.append('CMD=/usr/bin/zsh')
        changes.append('ENTRYPOINT=/bin/sh date')
        changes.append('ENV=TEST=test_containers.TestContainers.test_commit')
        changes.append('EXPOSE=80')
        changes.append('EXPOSE=8888')
        changes.append('LABEL=unittest=test_commit')
        changes.append('USER=bozo:circus')
        changes.append('VOLUME=/data')
        changes.append('WORKDIR=/data/application')

        id = self.alpine_ctnr.commit(
            'alpine3', author='Bozo the clown', changes=changes, pause=True)
        img = self.pclient.images.get(id)
        self.assertIsNotNone(img)

        details = img.inspect()
        self.assertEqual(details.author, 'Bozo the clown')
        self.assertListEqual(['/usr/bin/zsh'], details.containerconfig['cmd'])
        self.assertListEqual(['/bin/sh date'],
                             details.containerconfig['entrypoint'])
        self.assertIn('TEST=test_containers.TestContainers.test_commit',
                      details.containerconfig['env'])
        self.assertTrue(
            [e for e in details.containerconfig['env'] if 'PATH=' in e])
        self.assertDictEqual({
            '80': {},
            '8888': {},
        }, details.containerconfig['exposedports'])
        self.assertDictEqual({'unittest': 'test_commit'}, details.labels)
        self.assertEqual('bozo:circus', details.containerconfig['user'])
        self.assertEqual({'/data': {}}, details.containerconfig['volumes'])
        self.assertEqual('/data/application',
                         details.containerconfig['workingdir'])

    def test_remove(self):
        before = len(self.containers)

        with self.assertRaises(podman.ErrorOccurred):
            self.alpine_ctnr.remove()

        self.assertEqual(
            self.alpine_ctnr.id, self.alpine_ctnr.remove(force=True))
        self.loadCache()
        after = len(self.containers)

        self.assertLess(after, before)
        TestContainers.setUpClass()

    def test_restart(self):
        self.assertTrue(self.alpine_ctnr.running)
        before = self.alpine_ctnr.runningfor

        ctnr = self.alpine_ctnr.restart()
        self.assertTrue(ctnr.running)

        after = self.alpine_ctnr.runningfor

        # TODO: restore check when restart zeros counter
        # self.assertLess(after, before)

    def test_rename(self):
        with self.assertRaisesNotImplemented():
            self.alpine_ctnr.rename('new_alpine')

    def test_resize_tty(self):
        with self.assertRaisesNotImplemented():
            self.alpine_ctnr.resize_tty(132, 43)

    def test_pause_unpause(self):
        self.assertTrue(self.alpine_ctnr.running)

        ctnr = self.alpine_ctnr.pause()
        self.assertEqual(ctnr.status, 'paused')

        ctnr = self.alpine_ctnr.unpause()
        self.assertTrue(ctnr.running)
        self.assertTrue(ctnr.status, 'running')

    # creating cgoups can be flakey
    @retry(podman.libs.errors.ErrorOccurred, tries=4, delay=2, print_=print)
    def test_stats(self):
        try:
            self.assertTrue(self.alpine_ctnr.running)

            actual = self.alpine_ctnr.stats()
            self.assertEqual(self.alpine_ctnr.id, actual.id)
            self.assertEqual(self.alpine_ctnr.names, actual.name)
        except Exception:
            info = Path('/proc/self/mountinfo')
            with info.open() as fd:
                print('{} {}'.format(self.alpine_ctnr.id, info))
                print(fd.read())

    def test_logs(self):
        self.assertTrue(self.alpine_ctnr.running)
        actual = list(self.alpine_ctnr.logs())
        self.assertIsNotNone(actual)


if __name__ == '__main__':
    unittest.main()



"""Provide subparsers for pod commands."""
from pypodman.lib.actions.pod.create_parser import CreatePod
from pypodman.lib.actions.pod.inspect_parser import InspectPod
from pypodman.lib.actions.pod.kill_parser import KillPod
from pypodman.lib.actions.pod.pause_parser import PausePod
from pypodman.lib.actions.pod.processes_parser import ProcessesPod
from pypodman.lib.actions.pod.remove_parser import RemovePod
from pypodman.lib.actions.pod.start_parser import StartPod
from pypodman.lib.actions.pod.stop_parser import StopPod
from pypodman.lib.actions.pod.top_parser import TopPod
from pypodman.lib.actions.pod.unpause_parser import UnpausePod

__all__ = [
    'CreatePod',
    'InspectPod',
    'KillPod',
    'PausePod',
    'ProcessesPod',
    'RemovePod',
    'StartPod',
    'StopPod',
    'TopPod',
    'UnpausePod',
]



"""Decorator to retry failed method."""
import functools
import time


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, print_=None):
    """Retry calling the decorated function using an exponential backoff.

    Specialized for our unittests
    from:
    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    """
    def deco_retry(f):
        @functools.wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    if print_:
                        print_('{}, Retrying in {} seconds...'.format(
                            str(e), mdelay))
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry



"""Remote client command for run a command in a new container."""
import logging
import sys

import podman
from pypodman.lib import AbstractActionBase

from ._create_args import CreateArguments


class Run(AbstractActionBase):
    """Class for running a command in a container."""

    @classmethod
    def subparser(cls, parent):
        """Add Run command to parent parser."""
        parser = parent.add_parser('run', help='Run container from image')

        CreateArguments.add_arguments(parser)

        parser.add_argument('image', nargs=1, help='source image id.')
        parser.add_argument(
            'command',
            nargs='*',
            help='command and args to run.',
        )
        parser.set_defaults(class_=cls, method='run')

    def __init__(self, args):
        """Construct Run class."""
        super().__init__(args)
        if args.detach and args.rm:
            raise ValueError('Incompatible options: --detach and --rm')

        # image id used only on client
        del self.opts['image']

    def run(self):
        """Run container."""
        for ident in self._args.image:
            try:
                try:
                    img = self.client.images.get(ident)
                    ctnr = img.container(**self.opts)
                except podman.ImageNotFound as e:
                    sys.stdout.flush()
                    print(
                        'Image {} not found.'.format(e.name),
                        file=sys.stderr,
                        flush=True)
                    continue
                else:
                    logging.debug('New container created "{}"'.format(ctnr.id))

                if self._args.detach:
                    ctnr.start()
                    print(ctnr.id)
                else:
                    ctnr.attach(eot=4)
                    ctnr.start()
                    print(ctnr.id)

                    if self._args.rm:
                        ctnr.remove(force=True)
            except (BrokenPipeError, KeyboardInterrupt):
                print('\nContainer "{}" disconnected.'.format(ctnr.id))
            except podman.ErrorOccurred as e:
                sys.stdout.flush()
                print(
                    'Run for container "{}" failed: {} {}'.format(
                        ctnr.id, repr(e), e.reason.capitalize()),
                    file=sys.stderr,
                    flush=True)



"""Remote client command for deleting containers."""
import sys

import podman
from pypodman.lib import AbstractActionBase


class Rm(AbstractActionBase):
    """Class for removing containers from storage."""

    @classmethod
    def subparser(cls, parent):
        """Add Rm command to parent parser."""
        parser = parent.add_parser('rm', help='delete container(s)')
        parser.add_argument(
            '-f',
            '--force',
            action='store_true',
            help=('force delete of running container(s).'
                  ' (default: %(default)s)'))
        parser.add_argument(
            'targets', nargs='+', help='container id(s) to delete')
        parser.set_defaults(class_=cls, method='remove')

    def __init__(self, args):
        """Construct Rm class."""
        super().__init__(args)

    def remove(self):
        """Remove container(s)."""
        for ident in self._args.targets:
            try:
                ctnr = self.client.containers.get(ident)
                ctnr.remove(self._args.force)
                print(ident)
            except podman.ContainerNotFound as e:
                sys.stdout.flush()
                print(
                    'Container {} not found.'.format(e.name),
                    file=sys.stderr,
                    flush=True)
            except podman.ErrorOccurred as e:
                sys.stdout.flush()
                print(
                    '{}'.format(e.reason).capitalize(),
                    file=sys.stderr,
                    flush=True)



"""Implement common create container arguments together."""

from pypodman.lib import BooleanAction, UnitAction


class CreateArguments():
    """Helper to add all the create flags to a command."""

    @classmethod
    def add_arguments(cls, parser):
        """Add CreateArguments to parser."""
        parser.add_argument(
            '--add-host',
            action='append',
            metavar='HOST',
            help='Add a line to /etc/hosts.'
            ' The option can be set multiple times.'
            ' (format: hostname:ip)')
        parser.add_argument(
            '--annotation',
            action='append',
            help='Add an annotation to the container.'
            'The option can be set multiple times.'
            '(format: key=value)')
        parser.add_argument(
            '--attach',
            '-a',
            action='append',
            metavar='FD',
            help=('Attach to STDIN, STDOUT or STDERR. The option can be set'
                  ' for each of stdin, stdout, and stderr.'))
        parser.add_argument(
            '--blkio-weight',
            choices=range(10, 1000),
            metavar='[10-1000]',
            help=('Block IO weight (relative weight) accepts a'
                  ' weight value between 10 and 1000.'))
        parser.add_argument(
            '--blkio-weight-device',
            action='append',
            metavar='WEIGHT',
            help='Block IO weight, relative device weight.'
            ' (format: DEVICE_NAME:WEIGHT)')
        parser.add_argument(
            '--cap-add',
            action='append',
            metavar='CAP',
            help=('Add Linux capabilities'
                  'The option can be set multiple times.'))
        parser.add_argument(
            '--cap-drop',
            action='append',
            metavar='CAP',
            help=('Drop Linux capabilities'
                  'The option can be set multiple times.'))
        parser.add_argument(
            '--cgroup-parent',
            metavar='PATH',
            help='Path to cgroups under which the cgroup for the'
            ' container will be created. If the path is not'
            ' absolute, the path is considered to be relative'
            ' to the cgroups path of the init process. Cgroups'
            ' will be created if they do not already exist.')
        parser.add_argument(
            '--cidfile',
            metavar='PATH',
            help='Write the container ID to the file, on the remote host.')
        parser.add_argument(
            '--conmon-pidfile',
            metavar='PATH',
            help=('Write the pid of the conmon process to a file,'
                  ' on the remote host.'))
        parser.add_argument(
            '--cpu-period',
            type=int,
            metavar='PERIOD',
            help=('Limit the CPU CFS (Completely Fair Scheduler) period.'))
        parser.add_argument(
            '--cpu-quota',
            type=int,
            metavar='QUOTA',
            help=('Limit the CPU CFS (Completely Fair Scheduler) quota.'))
        parser.add_argument(
            '--cpu-rt-period',
            type=int,
            metavar='PERIOD',
            help=('Limit the CPU real-time period in microseconds.'))
        parser.add_argument(
            '--cpu-rt-runtime',
            type=int,
            metavar='LIMIT',
            help=('Limit the CPU real-time runtime in microseconds.'))
        parser.add_argument(
            '--cpu-shares',
            type=int,
            metavar='SHARES',
            help=('CPU shares (relative weight)'))
        parser.add_argument(
            '--cpus',
            type=float,
            help=('Number of CPUs. The default is 0.0 which means no limit'))
        parser.add_argument(
            '--cpuset-cpus',
            metavar='LIST',
            help=('CPUs in which to allow execution (0-3, 0,1)'))
        parser.add_argument(
            '--cpuset-mems',
            metavar='NODES',
            help=('Memory nodes (MEMs) in which to allow execution (0-3, 0,1).'
                  ' Only effective on NUMA systems'))
        parser.add_argument(
            '--detach',
            '-d',
            action=BooleanAction,
            default=False,
            help='Detached mode: run the container in the background and'
            ' print the new container ID. (default: False)')
        parser.add_argument(
            '--detach-keys',
            metavar='KEY(s)',
            default=4,
            help='Override the key sequence for detaching a container.'
            ' (format: a single character [a-Z] or ctrl-<value> where'
            ' <value> is one of: a-z, @, ^, [, , or _)')
        parser.add_argument(
            '--device',
            action='append',
            help=('Add a host device to the container'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--device-read-bps',
            action='append',
            metavar='LIMIT',
            help=('Limit read rate (bytes per second) from a device'
                  ' (e.g. --device-read-bps=/dev/sda:1mb)'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--device-read-iops',
            action='append',
            metavar='LIMIT',
            help=('Limit read rate (IO per second) from a device'
                  ' (e.g. --device-read-iops=/dev/sda:1000)'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--device-write-bps',
            action='append',
            metavar='LIMIT',
            help=('Limit write rate (bytes per second) to a device'
                  ' (e.g. --device-write-bps=/dev/sda:1mb)'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--device-write-iops',
            action='append',
            metavar='LIMIT',
            help=('Limit write rate (IO per second) to a device'
                  ' (e.g. --device-write-iops=/dev/sda:1000)'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--dns',
            action='append',
            metavar='SERVER',
            help=('Set custom DNS servers.'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--dns-option',
            action='append',
            metavar='OPT',
            help=('Set custom DNS options.'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--dns-search',
            action='append',
            metavar='DOMAIN',
            help=('Set custom DNS search domains.'
                  'The option can be set multiple times.'),
        )
        parser.add_argument(
            '--entrypoint',
            help=('Overwrite the default ENTRYPOINT of the image.'),
        )
        parser.add_argument(
            '--env',
            '-e',
            action='append',
            help=('Set environment variables.'),
        )
        parser.add_argument(
            '--env-file',
            help=('Read in a line delimited file of environment variables,'
                  ' on the remote host.'),
        )
        parser.add_argument(
            '--expose',
            action='append',
            metavar='RANGE',
            help=('Expose a port, or a range of ports'
                  ' (e.g. --expose=3300-3310) to set up port redirection.'),
        )
        parser.add_argument(
            '--gidmap',
            metavar='MAP',
            action='append',
            help=('GID map for the user namespace'),
        )
        parser.add_argument(
            '--group-add',
            action='append',
            metavar='GROUP',
            help=('Add additional groups to run as'))
        parser.add_argument('--hostname', help='Container host name')

        # only way for argparse to handle these options.
        vol_args = {
            'choices': ['bind', 'tmpfs', 'ignore'],
            'metavar': 'MODE',
            'type': str.lower,
            'help': 'Tells podman how to handle the builtin image volumes',
        }

        volume_group = parser.add_mutually_exclusive_group()
        volume_group.add_argument('--image-volume', **vol_args)
        volume_group.add_argument('--builtin-volume', **vol_args)

        parser.add_argument(
            '--interactive',
            '-i',
            action=BooleanAction,
            default=False,
            help='Keep STDIN open even if not attached. (default: False)')
        parser.add_argument('--ipc', help='Create namespace')
        parser.add_argument(
            '--kernel-memory', action=UnitAction, help='Kernel memory limit')
        parser.add_argument(
            '--label',
            '-l',
            action='append',
            help=('Add metadata to a container'
                  ' (e.g., --label com.example.key=value)'))
        parser.add_argument(
            '--label-file', help='Read in a line delimited file of labels')
        parser.add_argument(
            '--log-driver',
            choices='json-file',
            metavar='json-file',
            default='json-file',
            help='Logging driver for the container. (default: %(default)s)')
        parser.add_argument(
            '--log-opt',
            action='append',
            help='Logging driver specific options')
        parser.add_argument(
            '--memory', '-m', action=UnitAction, help='Memory limit')
        parser.add_argument(
            '--memory-reservation',
            action=UnitAction,
            help='Memory soft limit')
        parser.add_argument(
            '--memory-swap',
            action=UnitAction,
            help=('A limit value equal to memory plus swap.'
                  'Must be used with the --memory flag'))
        parser.add_argument(
            '--memory-swappiness',
            choices=range(0, 100),
            metavar='[0-100]',
            help="Tune a container's memory swappiness behavior")
        parser.add_argument('--name', help='Assign a name to the container')
        parser.add_argument(
            '--network',
            '--net',
            metavar='BRIDGE',
            help='Set the Network mode for the container.'
            ' (format: bridge, host, container:UUID, ns:PATH, none)')
        parser.add_argument(
            '--oom-kill-disable',
            action=BooleanAction,
            help='Whether to disable OOM Killer for the container or not')
        parser.add_argument(
            '--oom-score-adj',
            choices=range(-1000, 1000),
            metavar='[-1000-1000]',
            help="Tune the host's OOM preferences for containers")
        parser.add_argument(
            '--pid',
            help='Set the PID Namespace mode for the container.'
            '(format: host, container:UUID, ns:PATH)')
        parser.add_argument(
            '--pids-limit',
            type=int,
            metavar='LIMIT',
            help=("Tune the container's pids limit."
                  " Set -1 to have unlimited pids for the container."))
        parser.add_argument('--pod', help='Run container in an existing pod')
        parser.add_argument(
            '--privileged',
            action=BooleanAction,
            help='Give extended privileges to this container.')
        parser.add_argument(
            '--publish',
            '-p',
            metavar='RANGE',
            help="Publish a container's port, or range of ports, to the host")
        parser.add_argument(
            '--publish-all',
            '-P',
            action=BooleanAction,
            help='Publish all exposed ports to random'
            ' ports on the host interfaces'
            '(default: False)')
        parser.add_argument(
            '--quiet',
            '-q',
            action='store_true',
            help='Suppress output information when pulling images')
        parser.add_argument(
            '--read-only',
            action=BooleanAction,
            help="Mount the container's root filesystem as read only.")
        parser.add_argument(
            '--rm',
            action=BooleanAction,
            default=False,
            help='Automatically remove the container when it exits.')
        parser.add_argument(
            '--rootfs',
            action='store_true',
            help=('If specified, the first argument refers to an'
                  ' exploded container on the file system of remote host.'))
        parser.add_argument(
            '--security-opt',
            action='append',
            metavar='OPT',
            help='Set security options.')
        parser.add_argument(
            '--shm-size', action=UnitAction, help='Size of /dev/shm')
        parser.add_argument(
            '--sig-proxy',
            action=BooleanAction,
            default=True,
            help='Proxy signals sent to the podman run'
            ' command to the container process')
        parser.add_argument(
            '--stop-signal',
            metavar='SIGTERM',
            help='Signal to stop a container')
        parser.add_argument(
            '--stop-timeout',
            metavar='TIMEOUT',
            type=int,
            default=10,
            help='Seconds to wait on stopping container.')
        parser.add_argument(
            '--subgidname',
            metavar='MAP',
            help='Name for GID map from the /etc/subgid file')
        parser.add_argument(
            '--subuidname',
            metavar='MAP',
            help='Name for UID map from the /etc/subuid file')
        parser.add_argument(
            '--sysctl',
            action='append',
            help='Configure namespaced kernel parameters at runtime')
        parser.add_argument(
            '--tmpfs',
            action='append',
            metavar='MOUNT',
            help='Create a tmpfs mount.'
            ' (default: rw,noexec,nosuid,nodev,size=65536k.)')
        parser.add_argument(
            '--tty',
            '-t',
            action=BooleanAction,
            default=False,
            help='Allocate a pseudo-TTY for standard input of container.')
        parser.add_argument(
            '--uidmap',
            action='append',
            metavar='MAP',
            help='UID map for the user namespace')
        parser.add_argument(
            '--ulimit',
            action='append',
            metavar='OPT',
            help='Ulimit options',
        )
        parser.add_argument(
            '--user',
            '-u',
            help=('Sets the username or UID used and optionally'
                  ' the groupname or GID for the specified command.'))
        parser.add_argument(
            '--userns',
            metavar='NAMESPACE',
            help='Set the user namespace mode for the container')
        parser.add_argument(
            '--uts',
            choices=['host', 'ns'],
            help='Set the UTS mode for the container')
        parser.add_argument('--volume', '-v', help='Create a bind mount.')
        parser.add_argument(
            '--volumes-from',
            action='append',
            help='Mount volumes from the specified container(s).')
        parser.add_argument(
            '--workdir',
            '-w',
            metavar='PATH',
            help='Working directory inside the container')



import os
import unittest
from urllib.parse import urlparse

import podman
import varlink


class TestSystem(unittest.TestCase):
    def setUp(self):
        self.host = os.environ['PODMAN_HOST']
        self.tmpdir = os.environ['TMPDIR']

    def tearDown(self):
        pass

    def test_bad_address(self):
        with self.assertRaisesRegex(varlink.client.ConnectionError,
                                    "Invalid address 'bad address'"):
            podman.Client('bad address')

    def test_ping(self):
        with podman.Client(self.host) as pclient:
            self.assertTrue(pclient.system.ping())

    @unittest.skip("TODO: Need to setup ssh under Travis")
    def test_remote_ping(self):
        host = urlparse(self.host)
        remote_uri = 'ssh://root@localhost{}'.format(host.path)

        local_uri = 'unix:{}/tunnel/podman.sock'.format(self.tmpdir)
        with podman.Client(
                uri=local_uri,
                remote_uri=remote_uri,
                identity_file=os.path.expanduser('~/.ssh/id_rsa'),
        ) as remote_client:
            self.assertTrue(remote_client.system.ping())

    def test_versions(self):
        with podman.Client(self.host) as pclient:
            # Values change with each build so we cannot test too much
            self.assertListEqual(
                sorted([
                    'built', 'client_version', 'git_commit', 'go_version',
                    'os_arch', 'version'
                ]), sorted(list(pclient.system.versions._fields)))
            pclient.system.versions
        self.assertIsNot(podman.__version__, '0.0.0')

    def test_info(self):
        with podman.Client(self.host) as pclient:
            actual = pclient.system.info()
            # Values change too much to do exhaustive testing
            self.assertIsNotNone(actual.podman['go_version'])
            self.assertListEqual(
                sorted([
                    'host', 'insecure_registries', 'podman', 'registries',
                    'store'
                ]), sorted(list(actual._fields)))


if __name__ == '__main__':
    unittest.main()



"""Remote client command for starting pod and container(s)."""

import sys

import podman
from pypodman.lib import AbstractActionBase
from pypodman.lib import query_model as query_pods


class StartPod(AbstractActionBase):
    """Class for starting pod and container(s)."""

    @classmethod
    def subparser(cls, parent):
        """Add Pod Start command to parent parser."""
        parser = parent.add_parser('start', help='start pod')
        parser.add_argument(
            '-a', '--all', action='store_true', help='Start all pods')
        parser.add_argument(
            'pod', nargs='*', help='Pod to start. Or, use --all')
        parser.set_defaults(class_=cls, method='start')

    def __init__(self, args):
        """Construct StartPod object."""
        if args.all and args.pod:
            raise ValueError('You may give a pod or use --all, but not both')
        super().__init__(args)

    def start(self):
        """Start pod and container(s)."""
        idents = None if self._args.all else self._args.pod
        pods = query_pods(self.client.pods, idents)

        for pod in pods:
            try:
                pod.start()
            except podman.ErrorOccurred as ex:
                print(
                    '{}'.format(ex.reason).capitalize(),
                    file=sys.stderr,
                    flush=True)
                return 1
        return 0



"""Remote client command to import tarball as image filesystem."""
import sys

import podman
from pypodman.lib import AbstractActionBase


class Import(AbstractActionBase):
    """Class for importing tarball as image filesystem."""

    @classmethod
    def subparser(cls, parent):
        """Add Import command to parent parser."""
        parser = parent.add_parser(
            'import', help='import tarball as image filesystem')
        parser.add_argument(
            '--change',
            '-c',
            action='append',
            choices=('CMD', 'ENTRYPOINT', 'ENV', 'EXPOSE', 'LABEL',
                     'STOPSIGNAL', 'USER', 'VOLUME', 'WORKDIR'),
            type=str.upper,
            help='Apply the following possible instructions',
        )
        parser.add_argument(
            '--message', '-m', help='Set commit message for imported image.')
        parser.add_argument(
            'source',
            metavar='PATH',
            nargs=1,
            help='tarball to use as source on remote system',
        )
        parser.add_argument(
            'reference',
            metavar='TAG',
            nargs='*',
            help='Optional tag for image. (default: None)',
        )
        parser.set_defaults(class_=cls, method='import_')

    def __init__(self, args):
        """Construct Import class."""
        super().__init__(args)

    def import_(self):
        """Import tarball as image filesystem."""
        try:
            ident = self.client.images.import_image(
                self.opts.source,
                self.opts.reference,
                message=self.opts.message,
                changes=self.opts.change)
            print(ident)
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1



"""Remote client command for restarting containers."""
import logging
import sys

import podman
from pypodman.lib import AbstractActionBase, PositiveIntAction


class Restart(AbstractActionBase):
    """Class for Restarting containers."""

    @classmethod
    def subparser(cls, parent):
        """Add Restart command to parent parser."""
        parser = parent.add_parser('restart', help='restart container(s)')
        parser.add_argument(
            '--timeout',
            action=PositiveIntAction,
            default=10,
            help='Timeout to wait before forcibly stopping the container'
            ' (default: %(default)s seconds)')
        parser.add_argument(
            'targets', nargs='+', help='container id(s) to restart')
        parser.set_defaults(class_=cls, method='restart')

    def __init__(self, args):
        """Construct Restart class."""
        super().__init__(args)

    def restart(self):
        """Restart container(s)."""
        try:
            for ident in self._args.targets:
                try:
                    ctnr = self.client.containers.get(ident)
                    logging.debug('Restarting Container %s', ctnr.id)
                    ctnr.restart(timeout=self._args.timeout)
                    print(ident)
                except podman.ContainerNotFound as e:
                    sys.stdout.flush()
                    print(
                        'Container {} not found.'.format(e.name),
                        file=sys.stderr,
                        flush=True)
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)



"""Remote client command for reporting on Podman service."""
import json
import sys

import podman
import yaml
from pypodman.lib import AbstractActionBase


class Info(AbstractActionBase):
    """Class for reporting on Podman Service."""

    @classmethod
    def subparser(cls, parent):
        """Add Info command to parent parser."""
        parser = parent.add_parser(
            'info', help='report info on podman service')
        parser.add_argument(
            '--format',
            choices=('json', 'yaml'),
            help="Alter the output for a format like 'json' or 'yaml'."
            " (default: yaml)")
        parser.set_defaults(class_=cls, method='info')

    def __init__(self, args):
        """Construct Info class."""
        super().__init__(args)

    def info(self):
        """Report on Podman Service."""
        try:
            info = self.client.system.info()
        except podman.ErrorOccurred as e:
            sys.stdout.flush()
            print(
                '{}'.format(e.reason).capitalize(),
                file=sys.stderr,
                flush=True)
            return 1
        else:
            if self._args.format == 'json':
                print(json.dumps(info._asdict(), indent=2), flush=True)
            else:
                print(
                    yaml.dump(
                        dict(info._asdict()),
                        canonical=False,
                        default_flow_style=False),
                    flush=True)



"""Remote client commands dealing with images."""
import operator
from collections import OrderedDict

import humanize

import podman
from pypodman.lib import AbstractActionBase, Report, ReportColumn


class Images(AbstractActionBase):
    """Class for Image manipulation."""

    @classmethod
    def subparser(cls, parent):
        """Add Images commands to parent parser."""
        parser = parent.add_parser('images', help='list images')
        super().subparser(parser)
        parser.add_argument(
            '--sort',
            choices=['created', 'id', 'repository', 'size', 'tag'],
            default='created',
            type=str.lower,
            help=('Change sort ordered of displayed images.'
                  ' (default: %(default)s)'))

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--digests',
            action='store_true',
            help='Include digests with images. (default: %(default)s)')
        parser.set_defaults(class_=cls, method='list')

    def __init__(self, args):
        """Construct Images class."""
        super().__init__(args)

        self.columns = OrderedDict({
            'name':
            ReportColumn('name', 'REPOSITORY', 40),
            'tag':
            ReportColumn('tag', 'TAG', 10),
            'id':
            ReportColumn('id', 'IMAGE ID', 12),
            'created':
            ReportColumn('created', 'CREATED', 12),
            'size':
            ReportColumn('size', 'SIZE', 8),
            'repoDigests':
            ReportColumn('repoDigests', 'DIGESTS', 35),
        })

    def list(self):
        """List images."""
        images = sorted(
            self.client.images.list(),
            key=operator.attrgetter(self._args.sort))
        if not images:
            return

        rows = list()
        for image in images:
            fields = dict(image)
            fields.update({
                'created':
                humanize.naturaldate(podman.datetime_parse(image.created)),
                'size':
                humanize.naturalsize(int(image.size), binary=True),
                'repoDigests':
                ' '.join(image.repoDigests),
            })

            for r in image.repoTags:
                name, tag = r.split(':', 1)
                fields.update({
                    'name': name,
                    'tag': tag,
                })
                rows.append(fields)

        if not self._args.digests:
            del self.columns['repoDigests']

        with Report(self.columns, heading=self._args.heading) as report:
            report.layout(
                rows, self.columns.keys(), truncate=self._args.truncate)
            for row in rows:
                report.row(**row)



"""Report on pod's containers' processes."""
import operator
from collections import OrderedDict

from pypodman.lib import AbstractActionBase, Report, ReportColumn


class ProcessesPod(AbstractActionBase):
    """Report on Pod's processes."""

    @classmethod
    def subparser(cls, parent):
        """Add Images command to parent parser."""
        parser = parent.add_parser('ps', help='list processes of pod')
        super().subparser(parser)

        parser.add_argument(
            '--ctr-names',
            action='store_true',
            help='Include container name in the info field')
        parser.add_argument(
            '--ctr-ids',
            action='store_true',
            help='Include container ID in the info field')
        parser.add_argument(
            '--ctr-status',
            action='store_true',
            help='Include container status in the info field')
        parser.add_argument(
            '--format',
            choices=('json'),
            help='Pretty-print containers to JSON')
        parser.add_argument(
            '--sort',
            choices=('created', 'id', 'name', 'status', 'count'),
            default='created',
            type=str.lower,
            help='Sort on given field. (default: %(default)s)')
        parser.add_argument('--filter', help='Not Implemented')
        parser.set_defaults(class_=cls, method='processes')

    def __init__(self, args):
        """Contstruct ProcessesPod class."""
        if args.sort == 'created':
            args.sort = 'createdat'
        elif args.sort == 'count':
            args.sort = 'numberofcontainers'

        super().__init__(args)

        self.columns = OrderedDict({
            'id':
            ReportColumn('id', 'POD ID', 14),
            'name':
            ReportColumn('name', 'NAME', 30),
            'status':
            ReportColumn('status', 'STATUS', 8),
            'numberofcontainers':
            ReportColumn('numberofcontainers', 'NUMBER OF CONTAINERS', 0),
            'info':
            ReportColumn('info', 'CONTAINER INFO', 0),
        })

    def processes(self):
        """List pods."""
        pods = sorted(
            self.client.pods.list(), key=operator.attrgetter(self._args.sort))
        if not pods:
            return

        rows = list()
        for pod in pods:
            fields = dict(pod)
            if self._args.ctr_ids \
                    or self._args.ctr_names \
                    or self._args.ctr_status:
                keys = ('id', 'name', 'status', 'info')
                info = []
                for ctnr in pod.containersinfo:
                    ctnr_info = []
                    if self._args.ctr_ids:
                        ctnr_info.append(ctnr['id'])
                    if self._args.ctr_names:
                        ctnr_info.append(ctnr['name'])
                    if self._args.ctr_status:
                        ctnr_info.append(ctnr['status'])
                    info.append("[ {} ]".format(" ".join(ctnr_info)))
                fields.update({'info': " ".join(info)})
            else:
                keys = ('id', 'name', 'status', 'numberofcontainers')

            rows.append(fields)

        with Report(self.columns, heading=self._args.heading) as report:
            report.layout(rows, keys, truncate=self._args.truncate)
            for row in rows:
                report.row(**row)



"""Module to export all the podman subcommands."""
from pypodman.lib.actions.attach_action import Attach
from pypodman.lib.actions.commit_action import Commit
from pypodman.lib.actions.create_action import Create
from pypodman.lib.actions.export_action import Export
from pypodman.lib.actions.history_action import History
from pypodman.lib.actions.images_action import Images
from pypodman.lib.actions.import_action import Import
from pypodman.lib.actions.info_action import Info
from pypodman.lib.actions.inspect_action import Inspect
from pypodman.lib.actions.kill_action import Kill
from pypodman.lib.actions.logs_action import Logs
from pypodman.lib.actions.mount_action import Mount
from pypodman.lib.actions.pause_action import Pause
from pypodman.lib.actions.pod_action import Pod
from pypodman.lib.actions.port_action import Port
from pypodman.lib.actions.ps_action import Ps
from pypodman.lib.actions.pull_action import Pull
from pypodman.lib.actions.push_action import Push
from pypodman.lib.actions.restart_action import Restart
from pypodman.lib.actions.rm_action import Rm
from pypodman.lib.actions.rmi_action import Rmi
from pypodman.lib.actions.run_action import Run
from pypodman.lib.actions.search_action import Search

__all__ = [
    'Attach',
    'Commit',
    'Create',
    'Export',
    'History',
    'Images',
    'Import',
    'Info',
    'Inspect',
    'Kill',
    'Logs',
    'Mount',
    'Pause',
    'Pod',
    'Port',
    'Ps',
    'Pull',
    'Push',
    'Restart',
    'Rm',
    'Rmi',
    'Run',
    'Search',
]



"""Remote client command for pausing processes in pod."""
import sys

import podman
from pypodman.lib import AbstractActionBase
from pypodman.lib import query_model as query_pods


class PausePod(AbstractActionBase):
    """Class for pausing containers in pod."""

    @classmethod
    def subparser(cls, parent):
        """Add Pod Pause command to parent parser."""
        parser = parent.add_parser('pause', help='pause containers in pod')
        parser.add_argument(
            '-a', '--all', action='store_true', help='Pause all pods')
        parser.add_argument('pod', nargs='*', help='pod(s) to pause.')
        parser.set_defaults(class_=cls, method='pause')

    def __init__(self, args):
        """Construct Pod Pause object."""
        if args.all and args.pod:
            raise ValueError('You may give a pod or use --all, but not both')
        super().__init__(args)

    def pause(self):
        """Pause containers in provided Pod."""
        idents = None if self._args.all else self._args.pod
        pods = query_pods(self.client.pods, idents)

        for pod in pods:
            try:
                pod.pause()
                print(pod.id)
            except podman.PodNotFound as ex:
                print(
                    'Pod "{}" not found'.format(ex.name),
                    file=sys.stderr,
                    flush=True)
            except podman.ErrorOccurred as ex:
                print(
                    '{}'.format(ex.reason).capitalize(),
                    file=sys.stderr,
                    flush=True)
                return 1
        return 0
